(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[810],{5833:function(n,u,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/clusters",function(){return t(3450)}])},3450:function(n,u,t){"use strict";t.r(u),t.d(u,{default:function(){return _}});var c=t(5893),r=t(9037);function _(){return(0,c.jsx)(r.Cc,{})}}},function(n){n.O(0,[678,312,979,845,37,888,774,179],function(){return n(n.s=5833)}),_N_E=n.O()}]);